-- character state machine
-- called by init
-- this: characterGO
function initCharacter()
	-- this function is called when the characters is newly created
	-- set the time the character needs from getting up to getting to work
	this.set("workPrepareTime", tonumber(readConfig("Game", "workPrepareTime")));
	-- set the minimum time the character thinks between actions
	this.set("minThinkTime", tonumber(readConfig("Game", "minThinkTime")));
	-- set the maximal time the character thinks between actions
	this.set("maxThinkTime", tonumber(readConfig("Game", "maxThinkTime")));


	-- add all trigger categories the character knows
	this.addTriggerCategory(Trigger.CategoryFlag.SIGHT);
	this.addTriggerCategory(Trigger.CategoryFlag.SOUND);
	this.addTriggerCategory(Trigger.CategoryFlag.SMELL);
	this.addTriggerCategory(Trigger.CategoryFlag.QUEST);

end


function manageJealousy(jealousyFlagName, charA, charB, charC)

		local gos = getGameObjectServer();

--		breakpoint()


		local romanticValueB = charA.getLongTermRelationship(charB.getCharacterID(), REL_ROMANTIC)
		local romanticValueC = charA.getLongTermRelationship(charC.getCharacterID(), REL_ROMANTIC)
		
		local romanticTreshold = _G["TRIANGLE_REL_TRESHOLD_"..jealousyFlagName]	
		if (romanticValueB < romanticTreshold and romanticValueC < romanticTreshold) then return 0 end

		local intensity = 0
		local suspiciousActivityB = charB.getCurrentActivity(charC)
		local suspiciousActivityC = charC.getCurrentActivity(charB)
		if (not suspiciousActivityB) and (not suspiciousActivityC)  then return end
		
		local relationshipCategory = _G["TRIANGLE_REL_"..jealousyFlagName]
		local gravityB = 0
		if suspiciousActivityB then
			gravityB = suspiciousActivityB.getRelationshipInfluenceValue(relationshipCategory)
		end
		local gravityC = 0
		if suspiciousActivityC then
			gravityC = suspiciousActivityC.getRelationshipInfluenceValue(relationshipCategory)
		end
		
		local visibility = this.canSeeObject(charB, math.pi/2.0, MAX_LINE_OF_SIGHT)
		if not visibility then return 0 end
		
		local transgressionTresHold = _G["TRIANGLE_TRESHOLD_"..jealousyFlagName]
		local gravity = math.max(gravityB, gravityC)
		if gravity < transgressionTresHold then return end
		local transgressionScale = _G["TRIANGLE_SCALE_"..jealousyFlagName]
		gravity = (gravity - transgressionTresHold)/transgressionScale


		local jealousy = cond(romanticValueB >= romanticTreshold, 1, 0) + cond(romanticValueC >= romanticTreshold, 2, 0)
		
		if jealousy == 0 or jealousy == 3 then return 0 end
		
		local funPenalty = gravity * _G["TRIANGLE_FUN_PENALTY_"..jealousyFlagName]
		local comfortPenalty = gravity * _G["TRIANGLE_COMFORT_PENALTY_"..jealousyFlagName]

		local culprit = charB
		if jealousy == 2 then
			culprit = charC
		end
		charA.addRelationshipFlag(culprit.getCharacterID(), jealousyFlagName, gravity)
		culprit.addRelationshipFlag(charA.getCharacterID(), jealousyFlagName.."Inverse", gravity)
		
		charA.decCondition(NEED_FUN, funPenalty)
		charA.decCondition(NEED_COMFORT, comfortPenalty)
		culprit.decCondition(NEED_FUN, funPenalty)
		culprit.decCondition(NEED_COMFORT, comfortPenalty)

		return jealousy

end


beginStateMachine()

	onEnter(function(msg)
		-- preload animations

		--print("character::onEnter archtype: " .. getArchtypeSuffix(this));


--		------ test --------------------------------------------------------------
--		local testFilter = function(character, command, standardGO, data)
--			print("testFilter: char:"..character.getCharacterName().." command:"..command.." object:"..standardGO.getType().." data:"..data);
--			if command == "sit" and standardGO.hasBehavior("sofa") then
--				return true;
--			end
--		end
--
--		local filter = GameEventFilter();
--		filter.setCallback(testFilter);
--		this.addInteractionGameEventFilter(filter)
--		--------------------------------------------------------------------------

		-- walk
		local wso = this.walkSO;
		wso.preloadAnimation(archize(this, "stand") .. "01");
		wso.preloadAnimation(archize(this, "stand") .. "02");
		wso.preloadAnimation(archize(this, "stand") .. "03");
		wso.preloadAnimation(archize(this, "stand") .. "04");
		wso.preloadAnimation(archizeWalk(this, "stand-walkLFF-"));
		wso.preloadAnimation(archizeWalk(this, "walkRFF-walkLFF-"));
		wso.preloadAnimation(archizeWalk(this, "walkLFF-walkRFF-"));

		-- hand
		local hso = this.handSO;
		hso.setDefaultStatePath("hand.free");

		-- mimic
		local mso = this.mimicSO;
		mso.preloadAnimation("mimicEyeBlink");
		mso.preloadAnimation("mimicNeutral1");
		mso.preloadAnimation("mimicNeutral2");
		mso.preloadAnimation("mimicNeutral3");
		mso.preloadAnimation("mimicEyeClose");
		mso.preloadAnimation("mimicSmile1");
		mso.preloadAnimation("mimicSmile2");
		mso.preloadAnimation("mimicSmile3");
		mso.preloadAnimation("mimicSmile4");
		mso.preloadAnimation("mimicSmile5");
		mso.preloadAnimation("mimicAngry1");
		mso.preloadAnimation("mimicSad1");


	end)

	onMsg("init", function(msg)
		initCharacter();
	end )

	onMsg(ENABLE_CHANGED_MESSAGE, function(msg)
--		print(this.getCharacterName() .. " got ENABLE_CHANGED_MESSAGE: " .. tostring(this.getEnable()))

		printSOInfo(this.walkSO, "ENABLE_CHANGED_MESSAGE: " .. tostring(this.getEnable()))
		
		local wso = this.walkSO;
		if this.getEnable() then
			wso.setDefaultStatePath("charStand.stand");

			-- give NPCs a clean outfit on location change
			-- but not the flatmates!
			local isFlatmate = (this.getHome() == this.getGameObjectServer().getHome());
			if this.isNPC() and (not isFlatmate) then
				this.setOutfitDirtiness(0);
				fillGuestConditions(this)

				-- update the long term relationship if not done within the last 12 hours
				this.evaluateRelationshipLua();
			end

			-- add or remove guest behavior script
			local characterHome = this.getHome()
			local currentWorld = getGameObjectServer().getOutsideWorld()
			if (characterHome == currentWorld) and (currentWorld ~= UNDEFINED) then
				-- not a guest
				this.removeBehaviorScript("scriptGuest");
			elseif this.isNPC() then
				-- guest
				this.addBehaviorScript("scriptGuest");
			end

		else
			freeHands(this)		
			wso.setDefaultStatePath("charStand.disabled");
		end
--		wso.setTargetStatePath(wso.getDefaultStatePathString());

		sendMsg(msg.name, this.mimicSO)
	end )



	onMsg(SAVEGAME_LOADED_MESSAGE, function(msg)
		-- add all trigger categories the character knows
		-- work around until all savegames are new ones
		this.addTriggerCategory(Trigger.CategoryFlag.SIGHT);
		this.addTriggerCategory(Trigger.CategoryFlag.SOUND);
		this.addTriggerCategory(Trigger.CategoryFlag.SMELL);
		this.addTriggerCategory(Trigger.CategoryFlag.QUEST);

		-- fix old save game
		this.setEmoticon(EMOTICON_INLOVE)
		this.setEmoticon("")
		
		--fix old savegame
		if not this.isSwimming() then this.walkSO.resetWalkState() end
	end )

	onMsg("walkAway", function(msg)
--		if this.getUncompletedActivityQueueCount() == 0 then
		if this.getActivityQueueCount() == 0 then
			walkAway(1.0, 3.0, this)

--			local actionPoint = this.findPointNear(1.0, 3.0);
--			if actionPoint then
--				local walkEntry = buildWalkEntry(this, actionPoint)
--				this.pushActivityQueueEntry(walkEntry);
--			end
		end
	end )


	onMsg("buildMenu", function(msg)

		local currentCharacter = getStateObjectFromID(msg.sender);
--		print("character buildMenu " .. currentCharacter.getCharacterName().. "->" .. this.getCharacterName());
		-- build the pie menu
		local button ;

		if not this.isSwimming() then


			local romanceButton = addPieMenuButton("pm_romance", "");
			romanceButton.addIcon("guiIconRelationshipRomantic");

				button = romanceButton.addPieMenuButton("pm_talkRelation", "talk", "Relation");
				button.addDescription(ACTIVITY, "talkRelation");

				button = romanceButton.addPieMenuButton("pm_talkCompliment", "talk", "Compliment");
				button.addDescription(ACTIVITY, "talkCompliment");

-- is now "Romantic"
--				button = romanceButton.addPieMenuButton("pm_holdNice", "hold", "Nice");
--				button.addDescription(ACTIVITY, "holdNice");

				button = romanceButton.addPieMenuButton("pm_holdNice", "hold", "Romantic");
				button.addDescription(ACTIVITY, "holdRomantic");

--				button = romanceButton.addPieMenuButton("pm_kissFriendly", "kiss", "Friendly");
--				button.addDescription(ACTIVITY, "kissFriendly");

				button = romanceButton.addPieMenuButton("pm_kissFriendly", "kiss", "Cheek");
				button.addDescription(ACTIVITY, "kissCheek");

				button = romanceButton.addPieMenuButton("pm_kissRomantic", "kiss", "Romantic");
				button.addDescription(ACTIVITY, "kissRomantic");

--				button = romanceButton.addPieMenuButton("pm_kissIntimate", "kiss", "Intimate");
--				button.addDescription(ACTIVITY, "kissIntimate");

				button = romanceButton.addPieMenuButton("pm_holdBehind", "hold", "FromBehind");
				button.addDescription(ACTIVITY, "holdFromBehind");

				button = romanceButton.addPieMenuButton("pm_talkILoveYou", "talk", "ILoveYou");
				button.addDescription(ACTIVITY, "talkILoveYou");

--~ 				if (canMarry(currentCharacter, this) and storyActive(this.getGameObjectServer())) then
--~ 					button = romanceButton.addPieMenuButton("pm_talkMarryMe", "talkMarryMe", "");
--~ 					button.addIcon("guiIconRelationErotic");
--~ 					button.addIcon("guiIconRelationFriendship");
--~ 					button.addDescription(ACTIVITY, "talkILoveYou");
--~ 				end

				addInviteToObjectButtons(romanceButton, "romance")


		 	  local eroticButton = addPieMenuButton("pm_erotic", "");
				eroticButton.addIcon("guiIconLibido");

				button = eroticButton.addPieMenuButton("pm_talkFlirt", "talk", "Flirt");
				button.addDescription(ACTIVITY, "talkFlirt");

				button = eroticButton.addPieMenuButton("pm_kissSexy", "kiss", "Sexy");
				button.addDescription(ACTIVITY, "kissSexy");


				button = eroticButton.addPieMenuButton("pm_kissMakeOut", "kiss", "MakeOut");
				button.addDescription(ACTIVITY, "kissMakeOut");

				--button = eroticButton.addPieMenuButton("pm_holdBehind", "hold", "FromBehind");
				--button.addDescription(ACTIVITY, "holdFromBehind");

--				for behavior, data in INVITE_TO_OBJECT_DATA do
--					local object = getObjectsWithBehavior(behavior)[1]
--					if object then
--						button = eroticButton.addPieMenuButton("pm_invite" .. string.capitalize(behavior), "inviteObject", behavior);
--						button.addDescription(ACTIVITY, data["activity"]);
--					end
--				end

				addInviteToObjectButtons(eroticButton, "erotic")

				if (currentCharacter.isMale()) then

--					button = eroticButton.addPieMenuButton("pm_slapButt", "grab", "SlapButt");
--					button.addDescription(ACTIVITY, "holdBehind");
				end;


			local otherButton = addPieMenuButton("pm_other", "");
			otherButton.addIcon("guiIconSpass");

				button = otherButton.addPieMenuButton("pm_talkJoke", "talk", "Joke");
				button.addDescription(ACTIVITY, "talkJoke");

				button = otherButton.addPieMenuButton("pm_talkTease", "talk", "Tease");
				button.addDescription(ACTIVITY, "talkTease");

			if TEST_BUTTONS then
				button = otherButton.addPieMenuButton("pm_showCloth", "showCloth", "showCloth");
				button.addDescription(ACTIVITY, "showClothes");
			end
				button = otherButton.addPieMenuButton("pm_tickle", "tickle");
				button.addDescription(ACTIVITY, "tickle");



			local friendshipButton = addPieMenuButton("pm_friendship", "");
			friendshipButton.addIcon("guiIconRelationshipFriendship");

				button = friendshipButton.addPieMenuButton("pm_talkCasual", "talk", "Casual");
				button.addDescription(ACTIVITY, "talkCasual");

				button = friendshipButton.addPieMenuButton("pm_talkGossip", "talk", "Gossip");
				button.addDescription(ACTIVITY, "talkGossip");

				button = friendshipButton.addPieMenuButton("pm_talkComfort", "talk", "Comfort");
				button.addDescription(ACTIVITY, "talkComfort");

				button = friendshipButton.addPieMenuButton("pm_talkHobby", "talk", "Hobby");
				button.addDescription(ACTIVITY, "talkHobby");

				button = friendshipButton.addPieMenuButton("pm_talkArgue", "talk", "Argue");
				button.addDescription(ACTIVITY, "talkArgue");
				-- action is started from checkInjustice function
				button.addDescription(NO_AI, "true")
				-- Dance together button
				button = friendshipButton.addPieMenuButton("pm_twoDance", "twoDance", "twoDance");
				button.addDescription(ACTIVITY, "twoDancing");



			if this.isNPC() then

				if canAskToLeaveFlat(currentCharacter, this) then

					button = addPieMenuButton("pm_askToLeaveFlat", "kickOut", "askedToLeaveFlat");
					button.addDescription(ACTIVITY, "askedToLeaveFlat");
					button.addDescription(NO_AI, "true")

					button = addPieMenuButton("pm_forceToLeaveFlat", "kickOut", "forcedToLeaveFlat");
					button.addDescription(ACTIVITY, "forcedToLeaveFlat");
					button.addDescription(NO_AI, "true")

				end

				if canAskToLeaveRoom(currentCharacter, this) then

					button = addPieMenuButton("pm_askToLeaveRoom", "kickOut", "askedToLeaveRoom");
					button.addDescription(ACTIVITY, "askedToLeaveRoom");
					button.addDescription(NO_AI, "true")

--					button = addPieMenuButton("pm_forceToLeaveRoom", "kickOut", "forcedToLeaveRoom");
--					button.addDescription(ACTIVITY, "forcedToLeaveRoom");

				end

			end
			

			if TEST_BUTTONS then

				local testButton = addPieMenuButton("pm_test", "");

--					button = testButton.addPieMenuButton("pm_emoticon on", "emoticon", "4");
--					button.addDescription(DONTQUEUE, "true");
--					button = testButton.addPieMenuButton("pm_emoticon off", "emoticon", "");
--					button.addDescription(DONTQUEUE, "true");
--
--
--					-- greeting for testing
--					button = testButton.addPieMenuButton("pm_greetShakeHands", "greet", "ShakeHands");
--					button = testButton.addPieMenuButton("pm_greetShakeHandsFriendly", "greet", "ShakeHandsFriendly");
--					button = testButton.addPieMenuButton("pm_greetKiss", "greet", "Kiss");
--					button = testButton.addPieMenuButton("pm_greet", "greet", getGreetType(currentCharacter, this));
--
--					-- line of sigth
					button = testButton.addPieMenuButton("pm_canSee", "canSee");
					button.addDescription(DONTQUEUE, "true");

					local emoticonButton = testButton.addPieMenuButton("test_emoticons", "");
						button = emoticonButton.addPieMenuButton("NONE", "showEmoticon", "");
						button.addDescription(DONTQUEUE, "true");
						local emoticons = findGlobalsStartingWith("EMOTICON_", "string")
						for id, emoticon in emoticons do
							button = emoticonButton.addPieMenuButton(id, "showEmoticon", emoticon);
							button.addDescription(DONTQUEUE, "true");
							button.addDescription(STYLE, "img");
							button.addDescription(TEXTURE, emoticon);
						end

					-- test anim
					local animButton = testButton.addPieMenuButton("pm_testAnims", "");
					local anims = {"SofaHoldHandsMFActiveStart", "SofaHoldHandsMFPassiveStart"}
					for i=1, table.getn(anims) do
						local anim = anims[i];

						button = animButton.addPieMenuButton(anim, "playAnim", anim);
						button.addDescription(DONTQUEUE, "true");

						button = animButton.addPieMenuButton(anim .. "_mirror", "playAnim", anim .. "_mirror");
						button.addDescription(DONTQUEUE, "true");
					end

					local emoButton = testButton.addPieMenuButton("pm_emo", "");
						local emos = {  "Cry",
								"DontKnow",
								"FeelSorry",
								"Flirt",
								"Happy",
								"Hungry",
								"Hygiene",
								"Refuse",
								"SayHello",
								"Shy",
								"ShyNaked",
								"Sigh",
								"Stress",
								"SwingArms",
								"Think",
								"Toilet",
								"Wait",
								"Wonder",
								"Yawn",
								"YouSmell"  }
						for i, emo in emos do
							button = emoButton.addPieMenuButton(emo, "makeEmoPose", emo);
							button.addDescription(DONTQUEUE, "true");
						end


					button = testButton.addPieMenuButton("pm_printRelationship", "printRelationship");
					button.addDescription(DONTQUEUE, "true");

					button = testButton.addPieMenuButton("pm_incRelationship", "incRelationship", "1.0");
					button.addDescription(DONTQUEUE, "true");

					button = testButton.addPieMenuButton("pm_decRelationship", "incRelationship", "-1.0");
					button.addDescription(DONTQUEUE, "true");

					-- look at for testing
					button = testButton.addPieMenuButton("pm_mimicTalk", "mimicTalk");
					button.addDescription(DONTQUEUE, "true");

					button = testButton.addPieMenuButton("pm_mimicLook", "mimicLook");
					button.addDescription(DONTQUEUE, "true");

					button = testButton.addPieMenuButton("pm_emoYouSmell", "emoYouSmell");
					button.addDescription(DONTQUEUE, "true");

					button = testButton.addPieMenuButton("pm_testLookAt", "testLookAt");
					button.addDescription(DONTQUEUE, "true");

					button = testButton.addPieMenuButton("pm_mimicEyeToggle", "mimicEyeToggle");
					button.addDescription(USERACTION, "true");

					-- play trumpet at for testing
					button = testButton.addPieMenuButton("pm_playTrumpet", "playTrumpet");
					button.addDescription(ACTIVITY, "playTrumpet");

					-- applaude at for testing
					button = testButton.addPieMenuButton("pm_applaudeRestrained", "applaude", "Restrained");
					button.addDescription(ACTIVITY, "applaude");

					button = testButton.addPieMenuButton("pm_applaudeEnthusiastic", "applaude", "Enthusiastic");
					button.addDescription(ACTIVITY, "applaude");

					button = testButton.addPieMenuButton("pm_applaudePleased", "applaude", "Pleased");
					button.addDescription(ACTIVITY, "applaude");

					button = testButton.addPieMenuButton("pm_changeOutfit", "changeOutfit", "STANDARD");
					button.addDescription(DONTQUEUE, "true");

				--button = addPieMenuButton("pm_talkCutscene", "talkCutscene");

--				local talkButton = addPieMenuButton("pm_talkSentences", "");
--
--					local topics = table.getKeys(DIALOG["Talk"]);
--
--					for i=1, table.getn(topics) do
--						local topic = topics[i];
--						button = talkButton.addPieMenuButton("pm_" .. topic, "talk3", topic);
--					end

--				local talkButton2 = addPieMenuButton("pm_talkTopics", "");
--
--					topics = table.getKeys(DIALOG_TOPICS);
--
--					for i=1, table.getn(topics) do
--						local topic = topics[i];
--						button = talkButton2.addPieMenuButton("pm_talk" .. topic, "talk4", topic);
--					end

			end --TEST_BUTTONS


		end


		-- notify the hand state object
		local character = getStateObjectFromID(msg.sender);
		if character then sendMsg("buildMenu", character.handSO) end


	end )

	onMsg("buildMenu_Bed", function(msg)

		local button;
		local bed = getStateObjectFromID(msg.sender);

		-- the active char who send the buildMenu-msg to the passive char
		local currentCharacter = getStateObjectFromID(tonumber(msg.data))

		local charButton = bed.addPieMenuButton(translate("pm_with") .. " " .. this.getCharacterName() .. "...", "");

			-- bed inteactions --------------------------------------
			button = charButton.addPieMenuButton("pm_sexBed", "bedInteract", "sex", this.getUniqueID());
			button.addDescription(ACTIVITY, "sexBed");

			button = charButton.addPieMenuButton("pm_pettingBed", "bedInteract", "petting",  this.getUniqueID());
			button.addDescription(ACTIVITY, "pettingBed");

			button = charButton.addPieMenuButton("pm_kissBed", "bedInteract", "kiss", this.getUniqueID());
			button.addDescription(ACTIVITY, "kissBed");

			button = charButton.addPieMenuButton("pm_sleepHold", "bedInteract", "sleepHold", this.getUniqueID());
			button.addDescription(ACTIVITY, "sleepHold");

			if (not currentCharacter.walkSO.isInStateMachine("charInteract")) then
				button = charButton.addPieMenuButton("pm_roseBed", "bedInteract", "rose", this.getUniqueID());
				button.addDescription(ACTIVITY, "seduceRose");
			end

	end )

	onMsg("buildMenu_Deckchair", function(msg)

		local button;
		local deckchair = getStateObjectFromID(msg.sender);

		-- the active char who send the buildMenu-msg to the passive char
		local currentCharacter = getStateObjectFromID(tonumber(msg.data))

		local charButton = deckchair.addPieMenuButton(translate("pm_with") .. " " .. this.getCharacterName() .. "...", "");

			-- deckchair interactions --------------------------------------
			button = charButton.addPieMenuButton("pm_rubLotion", "deckchairInteract", "rubLotion", this.getUniqueID());
			button.addDescription(ACTIVITY, "rubLotion");

	end )

	onMsg("buildMenu_Whirlpool", function(msg)

		local button;
		local whirlpool = getStateObjectFromID(msg.sender);

		-- the active char who send the buildMenu-msg to the passive char
		local currentCharacter = getStateObjectFromID(tonumber(msg.data))

		local charButton = whirlpool.addPieMenuButton(translate("pm_with") .. " " .. this.getCharacterName() .. "...", "");

			-- whirlpool interactions --------------------------------------
			button = charButton.addPieMenuButton("pm_sexWhirlpool", "whirlpoolInteract", "sex", this.getUniqueID());
			button.addDescription(ACTIVITY, "sexWhirlpool");

	end )


	onMsg("buildMenu_Sofa", function(msg)

		local button;
		local sofa = getStateObjectFromID(msg.sender);

		local charButton = sofa.addPieMenuButton(translate("pm_with") .. " " .. this.getCharacterName() .. "...", "");

			-- sofa inteactions -------------------------------------
--			button = charButton.addPieMenuButton("pm_getIntimateSofa", "sofaInteract", "kiss", this.getUniqueID());
--			button.addDescription(ACTIVITY, "kissIntimate");

			button = charButton.addPieMenuButton("pm_holdHands", "sofaInteract", "holdHands", this.getUniqueID());
			button.addDescription(ACTIVITY, "holdHands");

			button = charButton.addPieMenuButton("pm_cuddle", "sofaInteract", "cuddle", this.getUniqueID());
			button.addDescription(ACTIVITY, "kissIntimate");

			button = charButton.addPieMenuButton("pm_pillowFight", "sofaInteract", "pillowFight", this.getUniqueID());
			button.addDescription(ACTIVITY, "pillowFight");
			---------------------------------------------------------
	end )

	onMsg("buildMenu_BoardGame", function(msg)
		-- the boardgame
		local boardgame = getStateObjectFromID(msg.sender);
		-- the active char who send the buildMenu-msg to the boardgame
		local currentCharacter = getStateObjectFromID(tonumber(msg.data))
		if currentCharacter and not hasBoardGameQueued(currentCharacter) then
			boardgame.clearPieMenu() -- prevent buttons from mulitple char appear in boardgame menu
			local boardgame = getStateObjectFromID(msg.sender);
			local button = boardgame.addPieMenuButton("pm_joinBoardgame", "joinBoardgame", "", this.getUniqueID());
			button.addDescription(ACTIVITY, "playBoardgame");
		end
	end )

	onMsg("buildMenu_Camera", function(msg)
		local camera = getStateObjectFromID(msg.sender);
		local button = camera.addPieMenuButton("pm_joinFoto", "joinFoto", "", this.getUniqueID());
		button.addDescription(ACTIVITY, "poseFoto");
	end )



	-- generic talk
	onMsg("talk", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print(msg.name .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("talk", character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("talkType", msg.data);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("talkType", msg.data);
			thisWsoContext.set("waitAnim", "emoWait");

			local charTargetState = "charStand/charWalkInteract/charInteract/charTalk.talk";
			local thisTargetState = "charStand/charWalkInteract/charInteract/charTalk.talk";

			startInteraction(character, this, charTargetState, thisTargetState, wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )




	-- cutscene talk
	onMsg("talkMarryMe", function(msg)

		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print(">>+>> onMsg talkMarryMe character " .. character.getCharacterName() .. " -> " .. getCharacterName());

		-- marry here
		instantAbort(character);

	end )



	-- generic talk
	onMsg("talkCutscene", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("talkCutscene " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("talk", character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("topic", msg.data);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("topic", msg.data);
			thisWsoContext.set("waitAnim", "emoWait");

			local targetState = "charStand/charWalkInteract/charInteract/charTalkCutscene.talk";

			startInteraction(character, this, targetState, targetState, wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )




	-- generic kiss
	onMsg("kiss", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("kiss " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("kiss", character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("kissType", msg.data);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("kissType", msg.data);
			thisWsoContext.set("waitAnim", "emoWait");

			local targetState;
			if msg.data == "Romantic" or msg.data == "Sexy" or msg.data == "MakeOut" then
				targetState = "charStand/charWalkInteract/charInteract/charKissLong.kiss";
			else
				targetState = "charStand/charWalkInteract/charInteract/charKiss.kiss";
			end

			startInteraction(character, this, targetState, targetState, wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )



	-- invite kiss
	onMsg("inviteObject", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		local interactionPoint = findInteractionPoint("kiss", character);
		if interactionPoint then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("waitAnim", "emoWait");
			wsoContext.set("objectBehavior", msg.data);

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("waitAnim", "emoWait");
			thisWsoContext.set("objectBehavior", msg.data);

			targetState = "charStand/charWalkInteract/charInteract/charInviteObject.invite";

			startInteraction(character, this, targetState, targetState, wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )



	-- generic talk
	onMsg("tickle", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print(msg.name .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("kiss", character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("waitAnim", "emoWait");

			local charTargetState = "charStand/charWalkInteract/charInteract/charTickle.tickle";
			local thisTargetState = "charStand/charWalkInteract/charInteract/charTickle.tickle";

			startInteraction(character, this, charTargetState, thisTargetState, wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )


	-- new
	onMsg("sofaInteract", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);
		-- get the sofa on which this character is sitting
--		local sofa = this.getObjectWithActivity("sit");

		print("sofaInteract " .. tostring(msg.sender));

		character.addWalkabiltyIgnored(this)
		local freeSofa, actionPoint1, actionPoint2 = findSofaForInteraction(this, character);
		character.removeAllWalkabiltyIgnored()

		if (freeSofa) then


			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPointNames", {actionPoint2.getName()});
			wsoContext.set("object", freeSofa);
			-- special context
			wsoContext.set("sofa", freeSofa);
			wsoContext.set("partner", thisWso);
			wsoContext.set("activePart", true);
			--wsoContext.set("walkIgnore", this);

			-- interaction context
			thisWsoContext.set("actionPointNames", {actionPoint1.getName()});
			thisWsoContext.set("object", freeSofa);
			-- special context
			thisWsoContext.set("sofa", freeSofa);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("activePart", false);

--			if msg.data == "kiss" then
--				local targetState = "charStand/charWalkInteract/charSofa/charSofaInteract/charSofaSitClose." .. msg.data;
--				startInteraction(character, this, targetState, targetState, wsoContext, thisWsoContext);
--			elseif msg.data == "cuddle" then

			if msg.data == "cuddle" then
				local targetState =  "charStand/charWalkInteract/charSofa/charSofaInteract/charInteract/charActive/charSofaHoldHands/charSofa" .. string.capitalize(msg.data) .. "." .. msg.data;
				startInteraction(character, this, targetState, string.replace(targetState, "Active", "Passive"), wsoContext, thisWsoContext);
			else
				local targetState =  "charStand/charWalkInteract/charSofa/charSofaInteract/charInteract/charActive/charSofa" .. string.capitalize(msg.data) .. "." .. msg.data;
				startInteraction(character, this, targetState, string.replace(targetState, "Active", "Passive"), wsoContext, thisWsoContext);
			end

		else
			print("getIntimate: no bed found");
			instantAbort(character, EMOTICON_NOSOFA, "emoThink")
		end


	end )




	onMsg("bedInteract", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);
		-- get the bed on wcjich this character is laying


		local freeBed, actionPoint1, actionPoint2 = findBedForInteraction(this, character);
		if (freeBed) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;


				printSOInfo(wso, "bedInteract1 " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);
				printSOInfo(thisWso, "bedInteract1 " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

				-- create state machine contexts
				local wsoContext =  StateMachineContext();
				local thisWsoContext =  StateMachineContext();

				-- interaction context
				wsoContext.set("actionPointNames", {actionPoint2.getName()});
				wsoContext.set("object", freeBed);
				-- special context
				wsoContext.set("bed", freeBed);
				wsoContext.set("partner", thisWso);
				wsoContext.set("activePart", true);


				-- interaction context
				thisWsoContext.set("actionPointNames", {actionPoint1.getName()});
				thisWsoContext.set("object", freeBed);
				-- special context
				thisWsoContext.set("bed", freeBed);
				thisWsoContext.set("partner", wso);
				thisWsoContext.set("activePart", false);

				local buttonText = "pm_" .. msg.data .. "Bed";

				targetState = "charStand/charWalkInteract/charBed/charInteract/charBedLayClose/charActive/charInteract/charBed" .. string.capitalize(msg.data) .. "." .. msg.data;
				startInteraction(character, this, targetState, string.replace(targetState, "Active", "Passive"), wsoContext, thisWsoContext);

				printSOInfo(wso, "bedInteract2 " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);
				printSOInfo(thisWso, "bedInteract2 " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);
		else
			print("getIntimate: no action point found of layDown1, layDown2");
			instantAbort(character, EMOTICON_NODOUBLEBED, "emoThink")
		end


	end )

	onMsg("whirlpoolInteract", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("whirlpoolInteract " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local freeWhirlpool, actionPoint1, actionPoint2 = findWhirlpoolForInteraction(this, character);
		if freeWhirlpool then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

				print("free double bed found");

				-- create state machine contexts
				local wsoContext =  StateMachineContext();
				local thisWsoContext =  StateMachineContext();

				-- interaction context
				wsoContext.set("actionPointNames", {actionPoint2.getName()});
				wsoContext.set("object", freeWhirlpool);
				-- special context
				wsoContext.set("pool", freeWhirlpool);
				wsoContext.set("partner", thisWso);
				wsoContext.set("activePart", true);


				-- interaction context
				thisWsoContext.set("actionPointNames", {actionPoint1.getName()});
				thisWsoContext.set("object", freeWhirlpool);
				-- special context
				thisWsoContext.set("pool", freeWhirlpool);
				thisWsoContext.set("partner", wso);
				thisWsoContext.set("activePart", false);

				local buttonText = "pm_" .. msg.data .. "Whirlpool";

				targetState = "charStand/charWalkInteract/charWhirlpool/charInteract/charActive/charWhirlpool" .. string.capitalize(msg.data) .. "." .. msg.data;
				startInteraction(character, this, targetState, string.replace(targetState, "Active", "Passive"), wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_CANNOT, "emoThink")
		end


	end )

	-- new
	onMsg("deckchairInteract", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("deckchairInteract " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local deckchair, sitActionPoint, layActionPoint = findDeckchairForInteraction(character, this);
		if (deckchair) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

				-- create state machine contexts
				local wsoContext =  StateMachineContext();
				local thisWsoContext =  StateMachineContext();

				-- interaction context
				wsoContext.set("actionPointNames", {sitActionPoint.getName()});
				wsoContext.set("object", deckchair);
				-- special context
				wsoContext.set("deckchair", deckchair);
				wsoContext.set("activePart", true);
				wsoContext.set("partner", thisWso);
				wsoContext.set("walkIgnore", this)
				wsoContext.set("waitAnim", "emoWait");

				-- interaction context
				thisWsoContext.set("actionPointNames", {layActionPoint.getName()});
				thisWsoContext.set("object", deckchair);
				-- special context
				thisWsoContext.set("deckchair", deckchair);
				thisWsoContext.set("activePart", false);
				thisWsoContext.set("partner", wso);

				local buttonText = "pm_" .. msg.data;

				local charTargetState = "charStand/charDeckchairInteract/charWalkInteract/charDeckchairSit/charDeckchairCream.cream";
				local thisTargetState = "charStand/charWalkInteract/charDeckchairLay/charDeckchairLayBelly/charDeckchairInteract/charDeckchairCream.cream";

				startInteraction(character, this, charTargetState, thisTargetState, wsoContext, thisWsoContext);

		else
			print("deckchairInteract: no deckchair found");
			instantAbort(character, EMOTICON_NOSEAT, "emoThink")
		end

	end )



	-- new
	onMsg("showCloth", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("showCloth " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("talk", character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("waitAnim", "emoWait");

			local buttonText = "pm_showCloth";

			local charTargetState = "charStand/charWalkInteract/charInteract/charShowCloth.show";
			local thisTargetState = "charStand/charInteract/charShowCloth.watch";

			startInteraction(character, this, charTargetState, thisTargetState, wsoContext, thisWsoContext);

		else
			print("deckchairInteract: no interaction");
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )

	-- Danser Ensemble
	onMsg("twoDance", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("twoDance" .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("talk", character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("waitAnim", "emoWait");

			local buttonText = "pm_twoDance";

			local charTargetState = "charStand/charWalkInteract/charInteract/charShowCloth.tdance";
			local thisTargetState = "charStand/charInteract/charShowCloth.tdance";

			startInteraction(character, this, charTargetState, thisTargetState, wsoContext, thisWsoContext);

		else
			print("deckchairInteract: no interaction");
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )

	-- new
	onMsg("applaude", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("applaude " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("talk", character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("applauseType", msg.data);
			wsoContext.set("partner", thisWso);
			wsoContext.set("waitAnim", "emoWait");

			local buttonText = "pm_applaude";

			local charTargetState = "charStand/charWalkInteract/charApplause.clap1";

			wso.setTargetStatePath(charTargetState, wsoContext);

--			-- if passive character is busy with somerthing else interrupt him
			if (passiveEntry ~= this.getActivityQueueEntry(0)) then this.cancelCurrentActivity() end;

		else
			print("deckchairInteract: no interaction");
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )

	-- generic hold
	onMsg("hold", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("hold " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local actionPointName = cond(msg.data == "FromBehind", "behind", "hug");
		local interactionPoint = findInteractionPoint(actionPointName, character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("holdType", msg.data);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("holdType", msg.data);
			thisWsoContext.set("waitAnim", "emoWait");

			local targetState;
			if msg.data == "Romantic" or msg.data == "FromBehind" then
				targetState = "charStand/charWalkInteract/charInteract/charHoldLong.hold";
			else
				targetState = "charStand/charWalkInteract/charInteract/charHold.hold";
			end
			startInteraction(character, this, targetState, targetState, wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )


	onMsg("grab", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("hold " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("behindSide1", character);
		if not interactionPoint then interactionPoint = findInteractionPoint("behindSide2", character); end

		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("grabType", msg.data);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("grabType", msg.data);
			thisWsoContext.set("waitAnim", "emoWait");

			local targetState = "charStand/charWalkInteract/charInteract/charActive/charGrab.grab"
			startInteraction(character, this, targetState, string.replace(targetState, "Active", "Passive"), wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )





	-- genetic hold
	onMsg("greet", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("hold " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint(cond(msg.data == "Kiss", "kiss", "talk"), character);

		if (interactionPoint) then

			-- for testing only
--			sendMsg("talk", this.mimicSO, tostring(msg.sender))
--			sendMsg("talk", character.mimicSO, tostring(this.getUniqueID()))

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("greetType", msg.data);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("greetType", msg.data);
			thisWsoContext.set("waitAnim", "emoWait");

			local targetState = "charStand/charWalkInteract/charInteract/charGreet.greet";

			startInteraction(character, this, targetState, targetState, wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )




	-- giveGift
	onMsg("giveGift", function(msg)

		-- get the game object server
		local gameObjectServer = getGameObjectServer();
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		print("giveGift " .. character.getCharacterName() .. " -> " .. this.getCharacterName() .. " data: " .. msg.data);

		local interactionPoint = findInteractionPoint("talk", character);
		if (interactionPoint) then

			-- get the walk state objects
			local wso = character.walkSO;
			local thisWso = this.walkSO;

			-- create state machine contexts
			local wsoContext =  StateMachineContext();
			local thisWsoContext =  StateMachineContext();

			-- interaction context
			wsoContext.set("actionPoint", interactionPoint.getActiveActionPoint());
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("activePart", true);
			wsoContext.set("partner", thisWso);
			wsoContext.set("waitAnim", "emoWait");

			-- interaction context
			thisWsoContext.set("actionPoint", interactionPoint.getPassiveActionPoint());
			thisWsoContext.set("object", character);
			-- special context
			thisWsoContext.set("activePart", false);
			thisWsoContext.set("partner", wso);
			thisWsoContext.set("waitAnim", "emoWait");

			local targetState = "charStand/charWalkInteract/charInteract/charGift.giveGift";

			startInteraction(character, this, targetState, targetState, wsoContext, thisWsoContext);

		else
			instantAbort(character, EMOTICON_PATH, "emoThink")
		end

	end )

---------------------------------------------------------------------------------------------------------------------------------------------
-- joinBoardgame


	onMsg("joinBoardgame", function(msg)

		-- send the character to chartableboardgame.join
		-- get the game object server
		local gameObjectServer = getGameObjectServer();
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		--if abortIfBroken(character, this) then return end;
		--if abortIfNotPossible(character, msg.name, this) then return end;

		local thisWso = this.walkSO;

		local tab = thisWso.get("table");
		local boardgame = thisWso.get("boardgame");
		if tab and boardgame then
			local tableActionPointName, chair = getFreeChairByTable(character, tab, "boardgameCards")
			if chair then
				-- get the walk state object
				local wso = character.walkSO;
				local wsoContext = StateMachineContext();
				-- store the action point
				wsoContext.set("actionPointNames", {"sit1", "sit2"});
				wsoContext.set("object", chair);
				-- special context
				wsoContext.set("partner", this.walkSO);
				wsoContext.set("chair", chair);
				wsoContext.set("boardgame", boardgame);
				wsoContext.set("table", tab);
				wsoContext.set("tableActionPointName", tableActionPointName);
				wsoContext.set("tableObject", "boardgameCards");

				wso.setTargetStatePath("charStand/charWalkInteract/charChair/charTablePlaceObject/charTableBoardGame.play", wsoContext);
				--character.getActivityQueueEntry(0).setGroupID(this.getActivityQueueEntry(0).getGroupID());

			else
				instantAbort(character, EMOTICON_NOSEAT, "emoThink")
			end
		else
			instantAbort(character, EMOTICON_CANNOT, "emoThink")
		end

	end )


---------------------------------------------------------------------------------------------------------------------------------------------
-- boardgameRequestHard

	onMsg("boardgameRequestHard", function(msg)
		local sender = getStateObjectFromID(msg.sender);
		print("boardgameRequestHard from ".. sender.getCharacterName() .." received by " .. this.getCharacterName())
		this.pushActivityQueueEntries({buildInteractionEntry(sender, "joinBoardgame", "", "pm_playBoardgame")}, true)
	end )

---------------------------------------------------------------------------------------------------------------------------------------------
-- boardgameRequest
-- trigger send by character waiting

	onMsg("boardgameRequest", function(msg)
		local sender = getStateObjectFromID(msg.sender);
		print("boardgameRequest from ".. sender.getCharacterName() .." received by " .. this.getCharacterName())

		this.addIgnoredTriggerMessage(msg.name, 30);

--		let the AI decide if the character should join the boardgame
		pushTargetCommand("joinBoardgame", sender, "");
		doSomethingNow(this);
	end )
---------------------------------------------------------------------------------------------------------------------------------------------

	onMsg("joinFoto", function(msg)

		-- get the game object server
		local gameObjectServer = getGameObjectServer();
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		local thisWso = this.walkSO;

		local camera = thisWso.get("camera");
		if camera then
			local poseActionPointsInSight = getActionPointsInSight(camera, "center", 1.7, camera, {"pose1","pose2","pose3"}, 1.7, character);

			-- get the walk state object
			local wso = character.walkSO;
			local wsoContext = StateMachineContext();
			-- store the action point
			wsoContext.set("actionPointNames", poseActionPointsInSight);
			wsoContext.set("object", camera);
			-- special context
			wsoContext.set("partner", this.walkSO);
			wsoContext.set("camera", camera);

			wso.setTargetStatePath("charStand/charWalkInteract/charCameraPose.pose", wsoContext);

		else
			instantAbort(character, EMOTICON_CANNOT, "emoThink")
		end

	end )

---------------------------------------------------------------------------------------------------------------------------------------------

	onMsg("fotoRequestHard", function(msg)
		local sender = getStateObjectFromID(msg.sender);
		print("fotogameRequestHard from ".. sender.getCharacterName() .." received by " .. this.getCharacterName())
		this.pushActivityQueueEntries({buildInteractionEntry(sender, "joinFoto", "", "pm_joinFoto")}, true)
	end )
---------------------------------------------------------------------------------------------------------------------------------------------

	onMsg("fotoRequest", function(msg)
		local sender = getStateObjectFromID(msg.sender);
		print("fotoRequest from ".. sender.getCharacterName() .." received by " .. this.getCharacterName())

		this.addIgnoredTriggerMessage(msg.name, 5);

--		let the AI decide if the character should join the boardgame
		pushTargetCommand("joinFoto", sender, "");
		doSomethingNow(this);

	end )

---------------------------------------------------------------------------------------------------------------------------------------------

	onMsg(OUTFIT_CHANGED_MESSAGE, function(msg)
		this.removeTrigger("naked");
		local dressCode = this.getDressCode();
		if (dressCode == UNDERWEAR or dressCode == NAKED) then
			print("OUTFIT_CHANGED_MESSAGE " .. getCharacterName() .. " create naked trigger")
			this.createTrigger(visualTriggerPrototype, "naked");
		end;

		updateAllLaundryGraphics(getGameObjectServer())
	end )

	-- trigger send by naked character
	onMsg("naked", function(msg)


		if readConfig("Debug", "shameNaked", "true") ~= "true" then return end;

		local sender = getStateObjectFromID(msg.sender);
		if sender == this then return end;

--		print("NACKED: " .. this.getCharacterName() .. " got naked from " .. sender.getCharacterName())

		if not this.isHere() then return end;
		if not sender.isHere() then return end;
		if this.getSleeping() then return end;
		if sender.getSleeping() then return end;
		if inBed(this) and not inBedInteraction(this) then return end;
		if inBed(sender) and not inBedInteraction(sender) then return end;

		local trigger = sender.findTrigger(msg.name);
		local dressCode = sender.getDressCode();
		local activityName = cond(dressCode == NAKED, "showNaked", "showInUnderwear");

		local possible = activityPossible(sender, activityName, this) or CharacterAI.checkFailing(ALWAYS_FAIL);

		if (not possible) then

			-- passive reaction
			if isIdle(this) then
				if not isOtherIntruder(this, sender) then
					if this.canSeeObject(sender, math.pi/2.0, MAX_LINE_OF_SIGHT) then
						if isOtherIntruder(sender, this) then
							local intruderEntries = {
								buildLeaveRoomEntry(this)
							}
							this.pushActivityQueueEntries(intruderEntries, true);
							this.setEmoticon(EMOTICON_SHY, EMOTICON_DELAY);
						else
--							this.cancelCurrentActivity();
							makeEmoPose(this, "emoShy", EMOTICON_SHY, sender, "back")
						end
					end
				end
			end

			if trigger and not hasOutfitChangeQueued(sender) then
			-- naked reaction
				if sender.canSeeObject(this, math.pi/2.0, MAX_LINE_OF_SIGHT) then
--					print("NACKED 2: naked " .. sender.getCharacterName() .. " can see " .. this.getCharacterName())

--					print("NACKED: naked " .. sender.getCharacterName() .. " remove naked trigger ")
--					sender.removeTrigger(msg.name);
					local nakedEntries = {
						buildEmoPoseEntry("emoShyNackedStart", EMOTICON_SHY, this,  "face"),
						buildDressUpEntry(sender),
--						buildMsgEntry(OUTFIT_CHANGED_MESSAGE)
					}
					sender.pushActivityQueueEntries(nakedEntries, true);
					--sender.cancelCurrentActivity();
				end
			end
		end
	end )
---------------------------------------------------------------------------------------------------------------------------------------------




	onMsg("eroticAction", function(msg)
--	seeing another character perform an erotic action may,
-- for now, only trigger an infidelity response

		local data = tonumber(msg.data)

		if data == nil then return end

		local gos = getGameObjectServer()

		local charB = getStateObjectFromID(msg.sender)
		if charB == this then return end

		local charC = gos.getCharacter(data)
		if charC == this then return end

		local jealousy = manageJealousy("infidelity", this, charB, charC)

		if (jealousy == 1 or jealousy == 3) then
			local argueEntry = buildInteractionEntry(charB, "talk", "Argue", "pm_talkArgue", "Normal")
			if argueEntry then this.queueActivityQueueEntry(argueEntry) end
		end
		if (jealousy == 2 or jealousy == 3) then
			local argueEntry = buildInteractionEntry(charC, "talk", "Argue", "pm_talkArgue", "Normal")
			if argueEntry then this.queueActivityQueueEntry(argueEntry) end
		end

		if (jealousy ~= 0) then
			setTopEntryStarted(this)
		end

	end )


	onMsg("romanticAction", function(msg)
--	seeing another character perform an erotic action may,
-- for now, only trigger a jealousy response

		local data = tonumber(msg.data)

		if data == nil then return end

		local gos = getGameObjectServer()

		local charB = getStateObjectFromID(msg.sender)
		if charB == this then return end

		local charC = gos.getCharacter(data)
		if charC == this then return end

		local jealousy = manageJealousy("jealousy", this, charB, charC)

		if (jealousy ~= 0) then
--			local leaveRoomActivity = buildLeaveRoomEntry(this)
--			if leaveRoomActivity then this.pushActivityQueueEntry(leaveRoomActivity) end
--			local emoEntry = buildEmoPoseEntry("emoCry", EMOTICON_SAD, charB)
--			if emoEntry then this.pushActivityQueueEntry(emoEntry) end
			this.pushActivityQueueEntries(
				{
					buildLeaveRoomEntry(this),
					buildEmoPoseEntry("emoCry", EMOTICON_SAD, charB)	
				}
			)
		end

	end )




--------------------------------------------------------------------------------------------------------------------------------------------------

	-- play as this char
	onMsg("playAsThis", function(msg)
		getGameObjectServer().setCurrentCharacter(this);
	end )

	-- visit character
	onMsg("visit", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender)
		local gos = getGameObjectServer();
		local gameWorld = gos.getOutsideWorld()
		local thisWorld = this.getOutsideWorld()
		if thisWorld ~= gameWorld then
		
			-- dress this character if not dressed appropriately
			dressForStreet(this);
		
			-- get the staircase or the bar dooe
			if gameWorld == BAR then
				-- place at the bar door
				placeNearObject(this, "barDoor", {"enter"})
				local wsoContext = StateMachineContext()
				wsoContext.set("inviter", character)
				this.walkSO.forceStatePath("charStand/charComeVisit/charOut.come", wsoContext)
			elseif table.contains(STAIRCASE_OUTSIDE_WORLDS, gameWorld) then
				-- place at the staircase
				local staircase = this.getObjectsWithBehavior("staircase")[1]
				if staircase then
					local wsoContext = StateMachineContext()
					wsoContext.set("staircase", staircase)
					wsoContext.set("inviter", character)
					this.walkSO.forceStatePath("charStand/charComeVisit/charWalkInteract/charStaircase/charOut.come", wsoContext)
				end
			end
			
--			local staircase = character.getObjectsWithBehavior("staircase")[1]
--			if staircase then
--				local wsoContext = StateMachineContext();
--				wsoContext.set("inviter", character)
--				wsoContext.set("staircase", staircase)
----				this.walkSO.setTargetStatePath("charStand/charStaircase/charOut.come", wsoContext);
--				this.walkSO.forceStatePath("charStand/charComeVisit/charStaircase/charOut.come", wsoContext);
--			else
--				print("visit: no staircase found")
--			end
		end
	end )

	-- play trumpet
	onMsg("playTrumpet", function(msg)
		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender)
		character.walkSO.setTargetStatePath("charStand/charTrumpetPlay.play");
	end )
	
	-- fix wrong previous locations
	onMsg(PRE_LOCATION_CHANGE_MESSAGE, function(msg)
		local gos = getGameObjectServer()
		local gameWorld = gos.getOutsideWorld()
		local thisWorld = this.getOutsideWorld()
		printSOInfo(this.walkSO, "PRE_LOCATION_CHANGE_MESSAGE enable:" .. tostring(this.getEnable()) .. " thisWorld:" .. tostring(thisWorld))		
		if thisWorld == gameWorld then -- character is staying in current world
			-- make sure current and previous world are the same
			this.setOutsideWorld(thisWorld)
		end
	end )

	onMsg(LOCATION_CHANGED_MESSAGE, function(msg)
		--breakpoint()
		--print(getCharacterName());
		--print(getOutsideWorld().toNumber());
		local gos = getGameObjectServer();

		local gameWorld = gos.getOutsideWorld()
		local thisWorld = this.getOutsideWorld()
		local thisPrevWorld = this.getPreviousOutsideWorld()
		
		-- assume character is not in new location
		this.setEnable(false)

		if gameWorld == thisWorld then
			-- character is in new location
			printSOInfo(this.walkSO, "LOCATION_CHANGED_MESSAGE enable:" .. tostring(this.getEnable()) .. " thisWorld:" .. tostring(thisWorld))		

			this.setEnable(true)

			if gameWorld ~= thisPrevWorld then
				-- has changed the location with the game
				if gameWorld == BAR then
					-- place at the bar door
					local dist = math.random(0, 3) * 0.5
					placeNearObject(this, "barDoor", nil, dist, dist + 2)
				elseif table.contains(STAIRCASE_OUTSIDE_WORLDS, gameWorld) then
					-- place at the staircase
					local staircase = this.getObjectsWithBehavior("staircase")[1]
					if staircase then
						local wsoContext = StateMachineContext()
						wsoContext.set("staircase", staircase)
--						this.walkSO.setTargetStatePath("charStand/charWalkInteract/charStaircase/charOut.come", wsoContext)
						this.walkSO.forceStatePath("charStand/charWalkInteract/charStaircase/charOut.come", wsoContext)
					else
						randomPlaceCharacter(this)
					end
				else
					randomPlaceCharacter(this)
				end
--				this.setPreviousOutsideWorld(gameWorld);
			else
				-- was already in new location
				fillGuestConditions(this) -- non guest get also filled
				randomPlaceCharacter(this)
				doSomethingNow(this)
			end

			-- give NPCs a clean outfit on location change
			if (this.isNPC()) then
				this.setOutfitDirtiness(0);
			end
		else
			-- character is not in new location
			this.setEnable(false)

			-- fix wrong character state if at bar but not in bar state
			if gameWorld == this.getHome() then
				if 
					not this.walkSO.isInStateMachine("charOut") and
					table.contains(STAIRCASE_OUTSIDE_WORLDS, gameWorld)
				then
					-- place at the staircase
					local staircase = this.getObjectsWithBehavior("staircase")[1]
					if staircase then
						local wsoContext = StateMachineContext()
						wsoContext.set("staircase", staircase)
						wsoContext.set("outDuration", BAR_STAY_TIME / 2);
						wsoContext.set("outsideWorld", thisWorld);
						this.walkSO.forceStatePath("charStand/charWalkInteract/charStaircase/charOut.go", wsoContext)
					end
				end
			else
				-- reset non present guest
				this.resetCharacter()
			end
			
		end
		
		-- make sure current and previous world are the same
		this.setOutsideWorld(thisWorld)

	end )


	onMsg("canSee", function(msg)
		local character = getStateObjectFromID(msg.sender)
		if character.canSeeObject(this, math.pi/2.0, MAX_LINE_OF_SIGHT) then
			character.setEmoticon(EMOTICON_SMILE, EMOTICON_DELAY);
		else
			character.setEmoticon(EMOTICON_CANNOT, EMOTICON_DELAY);
		end
	end )

	onMsg("printRelationship", function(msg)
		local character = getStateObjectFromID(msg.sender)
		printRelationship(character, this)
	end )

	onMsg("incRelationship", function(msg)
		local character = getStateObjectFromID(msg.sender)
		incRelationship(character, this, tonumber(msg.data))
	end )

	onMsg("mimicTalk", function(msg)
		local character = getStateObjectFromID(msg.sender)
		sendMsg("talk", character.mimicSO, tostring(this.getUniqueID()))
	end )

	onMsg("mimicLook", function(msg)
		local character = getStateObjectFromID(msg.sender)
		sendMsg("look", character.mimicSO, tostring(this.getUniqueID()))
	end )

	onMsg("mimicEyeToggle", function(msg)
		local mso = this.mimicSO;
		if mso.getState() == "eyeClose" then
			mso.setState("neutral1")
		else
			mso.setState("eyeClose")
		end
	end )

	onMsg("emoYouSmell", function(msg)
		local character = getStateObjectFromID(msg.sender)
		makeEmoPose(character, "emoYouSmell", EMOTICON_SMELL, this, "face")
	end )

	onMsg("testLookAt", function(msg)
		local talk = this.getActionPoint("talk")
		local behind = this.getActionPoint("behind")
		this.setLookAtActionPoint(talk, 1.5)
		this.pushActivityQueueEntry(buildWalkEntry(this, behind));
	end )

	onMsg("changeOutfit", function(msg)
		this.pushActivityQueueEntry(buildChangeOutfitEntry(this, msg.data));
	end )

	onMsg("doSomething", function(msg)
		doSomethingNow(this)
	end )

	onMsg("playAnim", function(msg)
		local wsoContext = StateMachineContext();
		wsoContext.set("animName", msg.data)
		wsoContext.set("loop", false)
		this.walkSO.setTargetStatePath("charStand/charPlayAnim.playAnim", wsoContext);
	end )

	onMsg("loopAnim", function(msg)
		local wsoContext = StateMachineContext();
		wsoContext.set("animName", msg.data)
		wsoContext.set("loop", true)
		this.walkSO.setTargetStatePath("charStand/charPlayAnim.playAnim", wsoContext);
	end )

	onMsg("makeEmoPose", function(msg)
		local context = StateMachineContext();
		local emoAnim = "emo" .. msg.data;
		makeEmoPose(this, emoAnim, emoticon);
	end )

	onMsg("showEmoticon", function(msg)
		this.setEmoticon(msg.data)
	end )

	onMsg("popupDialogStop", function(msg)
		sendMsg("stopAnim", this.walkSO);
	end )

	onMsg("emoticon", function(msg)
		local delay = tonumber(msg.data)
		if delay then
			setEmoticon(EMOTICON_INLOVE, delay)
		else
			setEmoticon("")
		end
	end )

	onMsg("clearQueue", function(msg)
		this.cancelAllActivities()
	end )

	onMsg("popupDialogStart", function(msg)

		if not this.getEnable() then
			local wsoContext = StateMachineContext();
			wsoContext.set("rightObjectName", "handleOldschool")
			wsoContext.set("animName", {genderize2(this, "phoneLoop"..1), genderize2(this, "phoneLoop"..2)})
			wsoContext.set("loop", true)
			this.walkSO.setTargetStatePath("charStand/charPlayAnim.playAnim", wsoContext);
		end

	end )

	onMsg("kickOut", function(msg)

		-- get character who initiated this action
		local character = getStateObjectFromID(msg.sender);

		if string.endsWith(msg.data, "Flat") and not canAskToLeaveFlat(character, this) then
			instantAbort(character, EMOTICON_CANNOT, "emoRefuseDezent")
			return
		end

		if string.endsWith(msg.data, "Room") and not canAskToLeaveRoom(character, this) then
			instantAbort(character, EMOTICON_CANNOT, "emoRefuseDezent")
			return
		end

		-- get the walk state object
		local wso = character.walkSO;
		local wsoContext = StateMachineContext()

		-- store the action point
		local params = FindPointNearParams();
--		params.actionPoint = this.getActionPoint("talk")
		params.actionPoint = this.getPositionAsActionPoint()
		params.minDist = 1.0
		params.maxDist = 2.5
		params.onlyTargetWalkable = true
		target = character.findPointNear(params);

		if target then
			target.setLookAt(params.actionPoint)

			wsoContext.set("actionPoint", target);
			wsoContext.set("object", this);
			-- special context
			wsoContext.set("kickOutType", msg.data);

			wso.setTargetStatePath("charStand/charWalkInteract/charKickOut.kickOut", wsoContext);
		else
			instantAbort(character, EMOTICON_CANNOT, "emoThink")
		end

	end )



endStateMachine()
